from __future__ import annotations

import base64
import hashlib
import hmac
from datetime import datetime
from typing import Any, Dict, Optional, TYPE_CHECKING

from pydantic import Extra, Field, BaseConfig, HttpUrl

from glQiwiApi.types.amount import HashablePlainAmount, PlainAmount
from glQiwiApi.types.base import HashableBase
from glQiwiApi.types.exceptions import WebhookSignatureUnverifiedError

if TYPE_CHECKING:
    from glQiwiApi.qiwi.clients.p2p.client import QiwiP2PClient  # noqa


class Customer(HashableBase):
    """Object: Customer"""

    phone_number: Optional[str] = Field(None, alias="phone")
    email: Optional[str] = None
    account: Optional[str] = None


class BillStatus(HashableBase):
    """Object: BillStatus"""

    value: str
    changed_datetime: Optional[datetime] = Field(None, alias="changedDateTime")


class CustomFields(HashableBase):
    """Object: CustomFields"""

    pay_sources_filter: Optional[str] = Field(None, alias="paySourcesFilter")
    theme_code: Optional[str] = Field(None, alias="themeCode")


class BillError(HashableBase):
    """Object: BillError"""

    service_name: str = Field(..., alias="serviceName")
    error_code: str = Field(..., alias="errorCode")
    description: str
    user_message: str = Field(..., alias="userMessage")
    datetime: str = Field(..., alias="dateTime")
    trace_id: str = Field(..., alias="traceId")


class Bill(HashableBase):
    """Object: Bill"""

    amount: HashablePlainAmount
    status: BillStatus
    site_id: str = Field(..., alias="siteId")
    id: str = Field(..., alias="billId")
    created_at: datetime = Field(..., alias="creationDateTime")
    expire_at: datetime = Field(..., alias="expirationDateTime")
    pay_url: str = Field(..., alias="payUrl")
    customer: Optional[Customer] = None
    custom_fields: Optional[CustomFields] = Field(None, alias="customFields")

    class Config(BaseConfig):
        extra = Extra.allow
        allow_mutation = True

    @property
    def invoice_uid(self) -> str:
        return self.pay_url[-36:]


class RefundedBill(HashableBase):
    """object: RefundedBill"""

    amount: PlainAmount
    datetime: datetime
    refund_id: str = Field(..., alias="refundId")
    status: str

    def __str__(self) -> str:
        return f"№{self.refund_id} {self.status} {self.amount} {self.datetime}"


class BillWebhookPayload(Bill):
    pay_url: None = Field(None, exclude=True)  # type: ignore


class BillWebhook(HashableBase):
    """Object: BillWebhook"""

    version: str = Field(..., alias="version")
    bill: BillWebhookPayload = Field(..., alias="bill")

    def __repr__(self) -> str:
        return f"#{self.bill.id} {self.bill.amount} {self.bill.status} "

    def verify_signature(self, sha256_signature: str, secret_p2p_key: str) -> None:
        webhook_key = base64.b64decode(bytes(secret_p2p_key, "utf-8"))
        bill = self.bill

        invoice_params = f"{bill.amount.currency}|{bill.amount.value}|{bill.id}|{bill.site_id}|{bill.status.value}"
        generated_signature = hmac.new(
            webhook_key, invoice_params.encode("utf-8"), hashlib.sha256
        ).hexdigest()

        if generated_signature != sha256_signature:
            raise WebhookSignatureUnverifiedError()


class PairOfP2PKeys(HashableBase):
    public_key: str = Field(..., alias="PublicKey")
    secret_key: str = Field(..., alias="SecretKey")


class InvoiceStatus(HashableBase):
    invoice_status: str
    is_sms_confirm: bool
    pay_results: Dict[Any, Any] = Field(..., alias="WALLET_ACCEPT_PAY_RESULT")


__all__ = (
    "Bill",
    "BillError",
    "RefundedBill",
    "BillWebhook",
    "PairOfP2PKeys",
    "InvoiceStatus",
    "BillStatus",
    "Customer",
)



from typing import ClassVar, Optional, Union

from pydantic import Field

from glQiwiApi.core.abc.api_method import APIMethod
from glQiwiApi.yoo_money.types import RequestPaymentResponse


class RequestPayment(APIMethod[RequestPaymentResponse]):
    http_method: ClassVar[str] = "POST"
    url: ClassVar[str] = "https://yoomoney.ru/api/request-payment"

    to_account: str = Field(..., alias="to")
    amount: Union[int, float] = Field(..., alias="amount_due")
    pattern_id: str = "p2p"
    comment_for_history: Optional[str] = Field(None, alias="comment")
    comment_for_receiver: Optional[str] = Field(None, alias="message")
    protect: bool = Field(False, alias="codepro")
    expire_period: int = 1



from datetime import datetime
from typing import Optional, List, Dict, Any, ClassVar

from pydantic import root_validator, conint, Field

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import History, TransactionType, Source
from glQiwiApi.utils.date_conversion import datetime_to_iso8601_with_moscow_timezone

MAX_HISTORY_LIMIT = 50


class GetHistory(QiwiAPIMethod[History]):
    http_method: ClassVar[str] = "GET"
    url: ClassVar[str] = "https://edge.qiwi.com/payment-history/v2/persons/{phone_number}/payments"

    rows: conint(le=MAX_HISTORY_LIMIT, strict=True, gt=0) = MAX_HISTORY_LIMIT
    transaction_type: TransactionType = Field(TransactionType.ALL, alias="type")
    sources: Optional[List[Source]] = Field(None)
    start_date: Optional[datetime] = Field(None, alias="startDate")
    end_date: Optional[datetime] = Field(None, alias="endDate")
    next_txn_date: Optional[datetime] = Field(None, alias="nextTxnDate")
    next_txn_id: Optional[int] = Field(None, alias="nextTxnId")

    class Config:
        use_enum_values = True

    @root_validator()
    def check_if_start_date_is_not_greater_than_end_date(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        start_date: Optional[datetime] = values.get("start_date")
        end_date: Optional[datetime] = values.get("end_date")

        both_are_not_transmitted = start_date is None and end_date is None

        if both_are_not_transmitted:
            return values

        only_start_date_transmitted = start_date is not None and end_date is None
        only_end_date_transmitted = start_date is None and end_date is not None

        if only_start_date_transmitted or only_end_date_transmitted:
            raise ValueError(
                "You must transmit both start_date and end_date parameters to get history rather than one"
            )

        if (end_date - start_date).total_seconds() <= 0:  # type: ignore
            raise ValueError("start_date cannot be bigger than end_date")

        return values

    @root_validator()
    def check_next_txn_date_and_next_txn_id(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        next_txn_id: Optional[int] = values.get("next_txn_id")
        next_txn_date: Optional[datetime] = values.get("next_txn_date")

        both_are_not_transmitted = next_txn_id is None and next_txn_date is None

        if both_are_not_transmitted:
            return values

        only_next_txn_date_transmitted = next_txn_date is not None and next_txn_id is None
        only_next_txn_id_transmitted = next_txn_date is None and next_txn_id is not None

        if only_next_txn_date_transmitted or only_next_txn_id_transmitted:
            raise ValueError(
                "You must transmit both next_txn_id and next_txn_date "
                "parameters to get history rather than one"
            )

        return values

    def build_request(self, **url_format_kw: Any) -> Request:
        data = self.dict(exclude={"sources"}, exclude_none=True, by_alias=True)

        if data.get("startDate") and data.get("endDate"):
            data["startDate"] = datetime_to_iso8601_with_moscow_timezone(data["startDate"])
            data["endDate"] = datetime_to_iso8601_with_moscow_timezone(data["endDate"])

        if data.get("nextTxnDate"):
            data["nextTxnDate"] = datetime_to_iso8601_with_moscow_timezone(data["nextTxnDate"])

        if self.sources is None:
            return Request(endpoint=self.url.format(**url_format_kw), params=data)

        for index, source in enumerate(self.sources, start=1):
            data.update({f"source[{index}]": source})

        return Request(endpoint=self.url.format(**url_format_kw), params=data)



from typing import ClassVar, Any

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import WebhookInfo


class RegisterWebhook(QiwiAPIMethod[WebhookInfo]):
    url: ClassVar[str] = "https://edge.qiwi.com/payment-notifier/v1/hooks"
    http_method: ClassVar[str] = "PUT"

    webhook_url: str = Field(..., alias="param")
    txn_type: int = Field(..., alias="txnType")

    hook_type: int = Field(default=1, alias="hookType")

    def build_request(self, **url_format_kw: Any) -> "Request":
        return Request(
            endpoint=self.url.format(**url_format_kw),
            params=self.dict(by_alias=True),
            http_method=self.http_method,
        )



from typing import Hashable

from pydantic import BaseModel, BaseConfig

from glQiwiApi.utils.compat import json


class Base(BaseModel):
    class Config(BaseConfig):
        json_dumps = json.dumps  # type: ignore
        json_loads = json.loads
        orm_mode = True


class HashableBase(Base):
    class Config(BaseConfig):
        allow_mutation = False

    def __hash__(self) -> int:
        return hash((type(self),) + tuple(self.__dict__.values()))

    def __eq__(self, other: Hashable) -> bool:
        return self.__hash__() == other.__hash__()



from typing import Optional, Dict, Any, Union

from pydantic import Field, validator

from glQiwiApi.types.amount import CurrencyModel, AmountWithCurrency
from glQiwiApi.types.base import HashableBase
from glQiwiApi.utils.currency_util import Currency


class AvailableBalance(HashableBase):
    alias: str
    currency: Union[str, CurrencyModel]


class Balance(HashableBase):
    """object: Balance"""

    alias: str
    title: str
    fs_alias: str = Field(alias="fsAlias")
    bank_alias: str = Field(alias="bankAlias")
    has_balance: bool = Field(alias="hasBalance")
    balance: Optional[AmountWithCurrency] = None
    currency: CurrencyModel
    account_type: Optional[Dict[str, Any]] = Field(None, alias="type")
    is_default_account: bool = Field(alias="defaultAccount")

    @validator("currency", pre=True)
    def humanize_pay_currency(cls, v):  # type: ignore
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))



from typing import ClassVar, Dict, Any

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.p2p.types import InvoiceStatus


class PayInvoice(QiwiAPIMethod[InvoiceStatus]):
    url: ClassVar[str] = "https://edge.qiwi.com/checkout-api/api/bill/search"
    http_method: ClassVar[str] = "POST"

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        "invoice_uid": RuntimeValue(),
        "currency": RuntimeValue(),
    }

    invoice_uid: str = Field(..., schema_path="invoice_uid")
    currency: str = Field(..., schema_path="currency")



from glQiwiApi.qiwi.clients.maps.client import QiwiMaps
from glQiwiApi.qiwi.clients.p2p.client import QiwiP2PClient
from glQiwiApi.qiwi.clients.wallet.client import QiwiWallet
from glQiwiApi.qiwi.clients.wrapper import QiwiWrapper
from .core.cache import InMemoryCacheStorage, APIResponsesCacheInvalidationStrategy
from .core.cache.storage import CacheStorage
from .yoo_money import YooMoneyAPI

try:
    import uvloop

    uvloop.install()  # pragma: no cover
except ImportError:
    pass


def default_cache_storage() -> CacheStorage:
    return InMemoryCacheStorage(invalidate_strategy=APIResponsesCacheInvalidationStrategy())


__version__ = "2.0.9"

__all__ = (
    # clients
    "YooMoneyAPI",
    "QiwiMaps",
    "QiwiWallet",
    "QiwiP2PClient",
    "QiwiWrapper",
    # other
    "__version__",
    "default_cache_storage",
)



from .abc import Pluggable
from .aiogram.polling import AiogramPollingPlugin
from .aiogram.webhook import AiogramWebhookPlugin

__all__ = ("Pluggable", "AiogramPollingPlugin", "AiogramWebhookPlugin")



from typing import ClassVar, Dict, Any, Union, Optional

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Commission


class PredictCommission(QiwiAPIMethod[Commission]):
    http_method: ClassVar[str] = "POST"
    url: ClassVar[str] = "https://edge.qiwi.com/sinap/providers/{private_card_id}/onlineCommission"

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        "account": RuntimeValue(),
        "paymentMethod": {"type": "Account", "accountId": "643"},
        "purchaseTotals": {"total": {"amount": RuntimeValue(), "currency": "643"}},
    }

    private_card_id: str = Field(..., path_runtime_value=True)

    invoice_amount: Union[str, int, float] = Field(..., scheme_path="purchaseTotals.total.amount")
    to_account: str = Field(..., scheme_path="account")



import time
from typing import ClassVar, Dict, Any

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types.qiwi_master import OrderDetails


class BuyQiwiMasterCard(QiwiAPIMethod[OrderDetails]):
    url: ClassVar[str] = "https://edge.qiwi.com/sinap/api/v2/terms/32064/payments"
    http_method: ClassVar[str] = "POST"

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        "id": RuntimeValue(default_factory=lambda: str(int(time.time() * 1000))),
        "sum": {"amount": RuntimeValue(default=99), "currency": "643"},
        "paymentMethod": {"type": "Account", "accountId": "643"},
        "fields": {"account": RuntimeValue(), "order_id": RuntimeValue()},
    }

    phone_number: str = Field(..., schema_path="fields.account")
    order_id: str = Field(..., schema_path="fields.order_id")



from pydantic import Field

from glQiwiApi.types.base import HashableBase


class Restriction(HashableBase):
    code: str = Field(..., alias="restrictionCode")
    description: str = Field(..., alias="restrictionDescription")


__all__ = ("Restriction",)



from pydantic import Field

from glQiwiApi.types.amount import AmountWithCurrency
from glQiwiApi.types.base import Base


class Commission(Base):
    provider_id: int = Field(alias="providerId")
    withdraw_sum: AmountWithCurrency = Field(alias="withdrawSum")
    enrollment_sum: AmountWithCurrency = Field(alias="enrollmentSum")
    qiwi_commission: AmountWithCurrency = Field(alias="qwCommission")
    withdraw_to_enrollment_rate: int = Field(alias="withdrawToEnrollmentRate")



from typing import TypeVar, NoReturn, Optional, List, Any


class YooMoneyError(Exception):
    def __init__(self, error_code: Optional[str] = None):
        super().__init__(error_code)


_T = TypeVar("_T")


def match_error(err_code: str) -> NoReturn:
    _MatchErrorMixin.detect(err_code)


class _MatchErrorMixin:
    match = ""
    explanation: str

    __subclasses: List[Any] = []

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super(_MatchErrorMixin, cls).__init_subclass__(**kwargs)
        if not hasattr(cls, f"_{cls.__name__}__group"):
            cls.__subclasses.append(cls)

    @classmethod
    def check(cls, message: str) -> bool:
        """
        Compare pattern with message
        :param message: always must be in lowercase
        :return: bool
        """
        return cls.match.lower() in message

    @classmethod
    def detect(cls, err_code: str) -> NoReturn:
        err_code = err_code.lower()
        for err in cls.__subclasses:
            if err is cls:
                continue
            if err.check(err_code):
                formatted_explanation = err.on_match(err_code, explanation=err.explanation)
                raise err(formatted_explanation or err_code)
        raise cls(err_code)  # type: ignore  # noqa

    @classmethod
    def on_match(cls, err_code: str, explanation: Optional[str] = None) -> Optional[str]:
        return cls.explanation


class BadRequest(YooMoneyError, _MatchErrorMixin):
    __group = True


class PaymentError(YooMoneyError, _MatchErrorMixin):
    __group = True


class IllegalParam(BadRequest):
    match = "illegal_param"
    explanation = "Invalid value for the {param} parameter."

    @classmethod
    def on_match(cls, err_code: str, explanation: Optional[str] = None) -> Optional[str]:
        param_name = err_code.removeprefix(cls.match)  # type: ignore
        param_name_without_underscore = param_name[1:]
        return explanation.format(param=param_name_without_underscore)  # type: ignore


class ProcessPaymentError(PaymentError):
    explanation = "Something went wrong while processing payment"


class ContractNotFound(ProcessPaymentError):
    match = "contract_not_found"
    explanation = "There is no existing unconfirmed payment with the specified request_id"


class NotEnoughFunds(ProcessPaymentError):
    match = "not_enough_funds"
    explanation = "The payer’s account does not have sufficient funds to make the payment. Additional funds should be credited to the account, and a new payment will need to be processed."


class LimitExceeded(ProcessPaymentError):
    match = "limit_exceeded"
    explanation = "One of the operation limits was exceeded:"
    "For the total amount of operations for the access token granted."
    "For the total amount of operations over a period of time for the access token granted."
    "YooMoney restrictions for various types of operations."


class MoneySourceNotAvailable(ProcessPaymentError):
    match = "money_source_not_available"
    explanation = "The requested payment method (money_source) is not available for this payment."


class PaymentRefused(ProcessPaymentError):
    match = "payment_refused"
    explanation = "The payment was refused. Possible reasons:"
    "The merchant refused to accept the payment (checkOrder request)."
    "The transfer to a YooMoney user is not possible (for example, the recipient’s wallet has reached the maximum amount allowed)."


class AuthorizationReject(ProcessPaymentError):
    match = "authorization_reject"
    explanation = """Authorization of the payment was refused. Possible reasons:

    The bank card expired.
    The issuing bank refused to perform the transaction for the card.
    Exceeds the limit for this user.
    A transaction with the current parameters is forbidden for this user.
    The user did not accept the User Agreement for the YooMoney service."""


class AccountBlocked(ProcessPaymentError):
    match = "account_blocked"
    explanation = """The user’s account has been blocked. In order to unblock the account, the user must be redirected to the address specified in the account_unblock_uri field."""



import time
from typing import ClassVar, Dict, Any, Optional

from pydantic import Field, validator

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import PaymentInfo


class TransferMoney(QiwiAPIMethod[PaymentInfo]):
    url: ClassVar[str] = "https://edge.qiwi.com/sinap/api/v2/terms/99/payments"
    http_method: ClassVar[str] = "POST"

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        "id": RuntimeValue(default_factory=lambda: str(int(time.time() * 1000))),
        "sum": {"amount": RuntimeValue(), "currency": "643"},
        "paymentMethod": {"type": "Account", "accountId": "643"},
        "comment": RuntimeValue(mandatory=False),
        "fields": {"account": RuntimeValue()},
    }

    @validator("to_wallet")
    def add_plus_sign_to_phone_number(cls, v: str) -> str:
        if v.startswith("+"):
            return v
        return f"+{v}"

    amount: int = Field(..., scheme_path="sum.amount")
    to_wallet: str = Field(..., scheme_path="fields.account")
    comment: Optional[str] = Field(None, scheme_path="comment")



from typing import Dict, ClassVar

from glQiwiApi.core.abc.api_method import APIMethod


class RejectIncomingTransfer(APIMethod[Dict[str, str]]):
    http_method: ClassVar[str] = "POST"
    url: ClassVar[str] = "https://yoomoney.ru/api/incoming-transfer-reject"

    operation_id: str



import uuid
from typing import ClassVar, Dict, Any, Optional

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import PaymentInfo, PaymentMethod, PaymentDetails
from glQiwiApi.types.amount import AmountWithCurrency


class MakePaymentByDetails(QiwiAPIMethod[PaymentInfo]):
    url: ClassVar[str] = "https://edge.qiwi.com/sinap/api/v2/terms/1717/payments"
    http_method: ClassVar[str] = "POST"

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        "id": RuntimeValue(default_factory=lambda: str(uuid.uuid4())),
        "sum": RuntimeValue(),
        "paymentMethod": RuntimeValue(),
        "fields": RuntimeValue(),
    }

    payment_sum: AmountWithCurrency = Field(..., scheme_path="sum")
    payment_method: PaymentMethod = Field(..., scheme_path="paymentMethod")
    details: PaymentDetails = Field(..., scheme_path="fields")
    payment_id: Optional[str] = Field(None, scheme_path="id")



from typing import Dict, Any, ClassVar

from glQiwiApi.qiwi.base import QiwiAPIMethod


class SendTestWebhookNotification(QiwiAPIMethod[Dict[Any, Any]]):
    http_method: ClassVar[str] = "GET"
    url: ClassVar[str] = "https://edge.qiwi.com/payment-notifier/v1/hooks/test"



from typing import Any, Dict, ClassVar

from pydantic import Field

from glQiwiApi.core.abc.api_method import APIMethod


class MakeCellularPayment(APIMethod[Dict[str, Any]]):
    http_method: ClassVar[str] = "POST"
    url: ClassVar[str] = "https://yoomoney.ru/api/operation-details"

    pattern_id: str
    phone_number: str = Field(..., alias="phone-number")
    amount: float



from glQiwiApi.core.event_fetching.webhooks.views.base import BaseWebhookView

from . import app  # noqa
from .views import QiwiBillWebhookView, QiwiTransactionWebhookView

__all__ = ("QiwiBillWebhookView", "QiwiTransactionWebhookView", "BaseWebhookView", "app.py")



import time
from typing import ClassVar, Dict, Any, Union

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue, Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import PaymentInfo


class TransferMoneyToCard(QiwiAPIMethod[PaymentInfo]):
    http_method: ClassVar[str] = "POST"
    url: ClassVar[str] = "https://edge.qiwi.com/sinap/api/v2/terms/{private_card_id}"

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        "id": RuntimeValue(default_factory=lambda: str(int(time.time() * 1000))),
        "sum": {"amount": RuntimeValue(), "currency": "643"},
        "paymentMethod": {"type": "Account", "accountId": "643"},
        "fields": {"account": RuntimeValue()},
    }

    private_card_id: str = Field(..., path_runtime_value=True)
    amount: Union[int, float] = Field(..., scheme_path="sum.amount")
    card_number: str = Field(..., scheme_path="fields.account")
    kwargs: Dict[str, Any] = {}  # parameters for cards with ID 1960, 21012

    def build_request(self, **url_format_kw: Any) -> "Request":
        request_schema = self._get_filled_json_payload_schema()
        request_schema["fields"].update(**self.kwargs)
        return Request(
            json_payload=request_schema,
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            http_method=self.http_method,
        )



from typing import Any, Dict, ClassVar, Optional

from pydantic import Field

from glQiwiApi.qiwi.base import QiwiAPIMethod


class AuthenticateWallet(QiwiAPIMethod[Dict[str, Any]]):
    url: ClassVar[
        str
    ] = "https://edge.qiwi.com/identification/v1/persons/{phone_number}/identification"
    http_method: ClassVar[str] = "POST"

    passport: str
    birth_date: str = Field(..., alias="birthDate")
    first_name: str = Field(..., alias="firstName")
    middle_name: str = Field(..., alias="middleName")
    last_name: str = Field(..., alias="lastName")
    inn: str = ""
    snils: str = ""
    oms: str = ""



import uuid
from datetime import datetime, timedelta
from typing import ClassVar, Dict, Any, Optional, Union, List

from pydantic import Field, validator

from glQiwiApi.core.abc.api_method import RuntimeValue, Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.p2p.types import Bill, Customer
from glQiwiApi.utils.date_conversion import datetime_to_iso8601_with_moscow_timezone


def get_default_bill_life_time() -> datetime:
    return datetime.now() + timedelta(days=2)


class CreateP2PBill(QiwiAPIMethod[Bill]):
    http_method: ClassVar[str] = "PUT"
    url: ClassVar[str] = "https://api.qiwi.com/partner/bill/v1/bills/{bill_id}"

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        "amount": {"currency": "RUB", "value": RuntimeValue()},
        "expirationDateTime": RuntimeValue(default_factory=get_default_bill_life_time),
        "comment": RuntimeValue(mandatory=False),
        "customFields": {
            "paySourcesFilter": RuntimeValue(default=["qw"]),
            "themeCode": RuntimeValue(default="Yvan-YKaSh"),
        },
        "customer": RuntimeValue(mandatory=False),
    }

    bill_id: Optional[str] = Field(None, path_runtime_value=True)
    expire_at: Optional[datetime] = Field(
        scheme_path="expirationDateTime",
    )
    amount: Union[int, float, str] = Field(..., scheme_path="amount.value")
    comment: Optional[str] = Field(None, scheme_path="comment")
    theme_code: Optional[str] = Field(None, scheme_path="customFields.themeCode")
    pay_source_filter: Optional[List[str]] = Field(
        None, scheme_path="customFields.paySourcesFilter"
    )
    customer: Optional[Customer] = Field(None, schema_path="customer")

    @validator("bill_id")
    def set_bill_id_if_it_is_none(cls, v: Optional[str]) -> str:
        return v or str(uuid.uuid4())

    def build_request(self, **url_format_kw: Any) -> "Request":
        request = super().build_request(**url_format_kw)
        expire_at = request.json_payload["expirationDateTime"]  # type: ignore
        request.json_payload["expirationDateTime"] = datetime_to_iso8601_with_moscow_timezone(
            expire_at
        )
        pay_source_filter = request.json_payload["customFields"]["paySourcesFilter"]
        if isinstance(pay_source_filter, list):
            request.json_payload["customFields"]["paySourcesFilter"] = " ".join(
                pay_source_filter
            ).replace(" ", ",")
        return request



from .abc.base_api_client import BaseAPIClient
from .event_fetching.class_based import (
    AbstractBillHandler,
    AbstractTransactionHandler,
    AbstractTransactionWebhookHandler,
    ErrorHandler,
    Handler,
)
from .event_fetching.filters import BaseFilter, LambdaBasedFilter
from .event_fetching.webhooks import (
    BaseWebhookView,
    QiwiBillWebhookView,
    QiwiTransactionWebhookView,
    app,
)
from .request_service import RequestService
from glQiwiApi.utils.synchronous import async_as_sync, execute_async_as_sync

__all__ = (
    "RequestService",
    "BaseFilter",
    "LambdaBasedFilter",
    # class-based handlers
    "Handler",
    "AbstractBillHandler",
    "AbstractTransactionHandler",
    "AbstractTransactionWebhookHandler",
    "ErrorHandler",
    # synchronous adapters and utils
    "async_as_sync",
    "execute_async_as_sync",
    # webhooks
    "QiwiBillWebhookView",
    "QiwiTransactionWebhookView",
    "BaseWebhookView",
)



import ipaddress
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import Field, validator

from glQiwiApi.types.amount import AmountWithCurrency, CurrencyModel
from glQiwiApi.types.base import Base
from glQiwiApi.utils.currency_util import Currency


class PassInfo(Base):
    """object: PassInfo"""

    last_pass_change: str = Field(alias="lastPassChange")
    next_pass_change: str = Field(alias="nextPassChange")
    password_used: bool = Field(alias="passwordUsed")


class MobilePinInfo(Base):
    """object: MobilePinInfo"""

    last_mobile_pin_change: Optional[datetime] = Field(alias="lastMobilePinChange")
    mobile_pin_used: bool = Field(alias="mobilePinUsed")
    next_mobile_pin_change: str = Field(alias="nextMobilePinChange")


class PinInfo(Base):
    """object: PinInfo"""

    pin_used: bool = Field(alias="pinUsed")


class AuthInfo(Base):
    """object: AuthInfo"""

    ip: ipaddress.IPv4Address
    bound_email: Optional[str] = Field(None, alias="boundEmail")
    last_login_date: Optional[datetime] = Field(None, alias="lastLoginDate")
    email_settings: Optional[Dict[Any, Any]] = Field(None, alias="emailSettings")
    mobile_pin_info: MobilePinInfo = Field(alias="mobilePinInfo")
    pass_info: PassInfo = Field(alias="passInfo")
    person_id: int = Field(alias="personId")
    pin_info: PinInfo = Field(alias="pinInfo")
    registration_date: datetime = Field(alias="registrationDate")


class SmsNotification(Base):
    """object: SmsNotification"""

    price: AmountWithCurrency
    enabled: bool
    active: bool
    end_date: Optional[datetime] = Field(None, alias="endDate")


class IdentificationInfo(Base):
    """object: IdentificationInfo"""

    bank_alias: Optional[str] = Field(None, alias="bankAlias")
    identification_level: Optional[str] = Field(None, alias="identificationLevel")
    passport_expired: Optional[bool] = Field(None, alias="passportExpired")


class NickName(Base):
    """object: NickName"""

    nickname: Optional[str] = None
    can_change: bool = Field(alias="canChange")
    can_use: bool = Field(alias="canUse")
    description: str = ""


class Feature(Base):
    """object: Feature"""

    feature_id: int = Field(alias="featureId")
    feature_value: str = Field(alias="featureValue")
    start_date: str = Field(alias="startDate")
    end_date: str = Field(alias="endDate")


class ContractInfo(Base):
    """object: ContractInfo"""

    blocked: bool = False
    contract_id: int = Field(alias="contractId")
    creation_date: datetime = Field(alias="creationDate")
    identification_info: List[IdentificationInfo] = Field(alias="identificationInfo")
    sms_notification: SmsNotification = Field(alias="smsNotification")
    nickname: NickName
    features: Optional[List[Feature]] = None


class UserInfo(Base):
    """object: UserInfo"""

    default_pay_currency: CurrencyModel = Field(..., alias="defaultPayCurrency")
    default_pay_source: Optional[int] = Field(None, alias="defaultPaySource")
    default_pay_account_alias: Optional[str] = Field(None, alias="defaultPayAccountAlias")
    email: Optional[str] = None
    first_transaction_id: Union[str, int, None] = Field(None, alias="firstTxnId")
    language: str
    operator: str
    phone_hash: str = Field(alias="phoneHash")
    promo_enabled: Optional[bool] = Field(None, alias="promoEnabled")

    @validator("default_pay_currency", pre=True)
    def humanize_pay_currency(cls, v):  # type: ignore
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))


class UserProfile(Base):
    auth_info: Optional[AuthInfo] = Field(None, alias="authInfo")
    contract_info: Optional[ContractInfo] = Field(None, alias="contractInfo")
    user_info: Optional[UserInfo] = Field(None, alias="userInfo")


__all__ = ["UserProfile"]



from pydantic import Field, BaseModel


class Polygon(BaseModel):
    """Polygon class for QiwiMaps class"""

    latitude_north_western: float = Field(..., alias="latNW")
    longitude_north_western: float = Field(..., alias="lngNW")
    latitude_south_east: float = Field(..., alias="latSE")
    longitude_south_east: float = Field(..., alias="lngSE")



import re
from typing import List, Any, cast, ClassVar

from glQiwiApi.core.abc.api_method import APIMethod, Request, ReturningType
from glQiwiApi.core.session.holder import HTTPResponse

YOO_MONEY_LINK_REGEXP = re.compile(r"https://yoomoney.ru/oauth2/authorize[?]requestid[=]\w+")


class BuildAuthURL(APIMethod[str]):
    http_method: ClassVar[str] = "POST"
    url: ClassVar[str] = "https://yoomoney.ru/oauth/authorize"

    client_id: str
    scopes: List[str]
    redirect_uri: str
    response_type: str = "code"

    def build_request(self, **url_format_kw: Any) -> "Request":
        return Request(
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            http_method=self.http_method,
            data={
                "client_id": self.client_id,
                "response_type": "code",
                "redirect_uri": self.redirect_uri,
                "scope": " ".join(self.scopes),
            }
        )

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> ReturningType:
        try:
            return cast(str, re.findall(YOO_MONEY_LINK_REGEXP, response.body.decode("utf-8"))[0])  # pragma: no cover
        except IndexError:
            raise Exception(
                "Could not find the authorization link in the response from "
                "the api, check the client_id value"
            )
